from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Callable, Optional, List, Any

from wwwpy.common.collectionlib import ListMap


@dataclass(frozen=True)
class Help:
    description: str
    url: str


_empty_help = Help("", "")


@dataclass
class NameHelp:
    name: str
    help: Help = field(default=_empty_help)

    @property
    def python_name(self) -> str:
        return self.name.replace('-', '_')


@dataclass
class EventDef(NameHelp):
    """Definition of an event of an HTML element."""
    pass




@dataclass
class AttributeDef(NameHelp):
    """Definition of an attribute of an HTML element."""
    values: list[str] = field(default_factory=list)
    boolean: bool = False
    mandatory: bool = False
    default_value: Optional[str] = None


class NamedListMap(ListMap):
    def __init__(self, args):
        super().__init__(args, key_func=lambda x: x.name)


@dataclass
class ElementDef:
    tag_name: str
    python_type: str
    help: Help = field(default=_empty_help)
    gen_html: Optional[Callable[[ElementDef, str], str]] = None
    """A function that generates the HTML for the element. It takes the data-name of the element as argument."""

    attributes: NamedListMap[AttributeDef] = field(default_factory=list)
    events: NamedListMap[EventDef] = field(default_factory=list)

    def __post_init__(self):
        if not isinstance(self.attributes, ListMap):
            self.attributes = NamedListMap(self.attributes)
        if not isinstance(self.events, ListMap):
            self.events = NamedListMap(self.events)

    def new_html(self, data_name: str) -> str:
        gen_html = self.gen_html or ElementDef.default_gen_html
        return gen_html(self, data_name)

    @classmethod
    def default_gen_html(cls, element_def: ElementDef, data_name: str) -> str:
        return f'\n<{element_def.tag_name} data-name="{data_name}"></{element_def.tag_name}>'


class ElementLibrary:
    def __init__(self):
        self.elements: List[ElementDef] = []

    def by_tag_name(self, tag_name: str) -> Optional[ElementDef]:
        for element in self.elements:
            if element.tag_name == tag_name:
                return element
        return None


_element_library: ElementLibrary = None


def element_library() -> ElementLibrary:
    global _element_library
    if _element_library is None:
        _element_library = ElementLibrary()
        from wwwpy.common.designer.shoelace import _shoelace_elements_def
        _element_library.elements.extend(_shoelace_elements_def())
        from .el_standard import _standard_elements_def
        _element_library.elements.extend(_standard_elements_def())
    return _element_library



from js import document


async def main():
    from wwwpy.remote import shoelace
    shoelace.setup_shoelace()

    from . import component1  # for component registration
    from . import upload_component  # for component registration
    document.body.innerHTML = '<component-1></component-1>'



from __future__ import annotations

import urllib
import urllib.request
from typing import Any

from wwwpy.http import HttpResponse


async def async_fetch_str(url: str, method: str = 'GET', data: str = '') -> str:
    response = sync_fetch_response(url, method=method, data=data)
    assert isinstance(response.content, str)
    return response.content


def sync_fetch_response(url: str, method: str = 'GET', data: str | bytes = '') -> HttpResponse:
    def make_response(r: Any) -> HttpResponse:
        return HttpResponse(
            r.read().decode("utf-8"),
            r.headers.get_content_type()
        )

    if method != 'GET':
        if isinstance(data, str):
            data_bytes = bytes(data, 'utf8')
        else:
            data_bytes = data

        rq = urllib.request.Request(url, method=method, data=data_bytes)
        with urllib.request.urlopen(rq) as r:
            return make_response(r)
    else:
        with urllib.request.urlopen(url) as r:
            return make_response(r)



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Union

import js
from js import console
from pyodide.ffi import create_proxy

import wwwpy.remote.component as wpc
from wwwpy.remote import dict_to_js
from wwwpy.remote.designer.global_interceptor import InterceptorEvent, GlobalInterceptor


@dataclass
class Actions:
    set_input_value = True
    # hide_dropdown = True
    # dispatch_change_event = True


class Option:
    parent: SearchableComboBox

    def __init__(self, text: str = ''):
        self.text = text
        self.actions = Actions()
        self.on_selected = lambda: None
        self._root_element: js.HTMLElement = None

    @property
    def loaded(self) -> bool:
        return self._root_element is not None

    @property
    def label(self) -> str:
        return self.root_element().innerText

    @label.setter
    def label(self, value: str):
        self.root_element().innerText = value

    @property
    def italic(self) -> bool:
        return self.root_element().style.fontStyle == 'italic'

    @italic.setter
    def italic(self, value: bool):
        self.root_element().style.fontStyle = 'italic' if value else ''

    def root_element(self) -> js.HTMLElement:
        if self._root_element is None:
            div = js.document.createElement('div')
            div.textContent = self.text
            div.addEventListener('click', create_proxy(self.do_click))
            self._root_element = div
        return self._root_element

    def do_click(self, *event):
        self.parent._option_selected(self)
        self.on_selected()

    def update_visibility(self, search: str):
        self.visible = search.lower() in self.text.lower()

    @property
    def visible(self) -> bool:
        return not (self._root_element.style.display == 'none')

    @visible.setter
    def visible(self, value):
        self.root_element().style.display = 'block' if value else 'none'


class OptionPopup(wpc.Component, tag_name='wwwpy-searchable-combobox2-option-popup'):
    _options = []  # type: List[Option]
    parent: SearchableComboBox
    _root: js.HTMLElement = wpc.element()
    _search: js.HTMLInputElement = wpc.element()
    _dirty = True

    def init_component(self):
        # language=html
        self.element.innerHTML = """
        <div><input type="search" data-name="_search">
        <div data-name="_root" class="dropdown"></div>
        </div>
        """
        self._interceptor = GlobalInterceptor(self._global_click)

    def _global_click(self, event: InterceptorEvent):
        target = event.event.composedPath()[0]
        click_inside = self.parent.root_element().contains(target)
        js.console.log(f'global click: contains={click_inside}', target)
        if click_inside:
            return
        self.hide()

    @property
    def visible(self) -> bool:
        return self.element.style.display != 'none'

    @property
    def options(self) -> List[Option]:
        return self._options

    @options.setter
    def options(self, value: List[Union[str, Option]]):
        def process(v):
            if not isinstance(v, Option):
                v = Option(v)
            v.parent = self.parent
            return v

        self._options = [process(v) for v in value]
        self._dirty = True

    @property
    def search_placeholder(self) -> str:
        return self._search.placeholder

    @search_placeholder.setter
    def search_placeholder(self, value):
        self._search.placeholder = value

    @property
    def search_value(self):
        return self._search.value

    @search_value.setter
    def search_value(self, value: str):
        self._search.value = value
        self._update_options_vis()

    def _search__input(self, event):
        self._update_options_vis()

    def _update_options_vis(self):
        for option in self._options:
            option.update_visibility(self._search.value)

    def show(self):
        if self._dirty:
            self._dirty = False
            root = self._root
            root.innerHTML = ''
            for option in self._options:
                option.parent = self.parent
                root.append(option.root_element())

        self.element.style.display = 'block'
        self._interceptor.install()

    def hide(self):
        self.element.style.display = 'none'
        self._interceptor.uninstall()

    def focus_search(self):
        self._search.focus()


class SearchableComboBox(wpc.Component, tag_name='wwwpy-searchable-combobox2'):
    _input: js.HTMLInputElement = wpc.element()
    option_popup: OptionPopup = wpc.element()
    focus_search_on_popup = True
    """This represents the popoup with the options and eventually a search box at the top to filter the options"""

    def root_element(self):
        return self.shadow

    def init_component(self):
        self.shadow = self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.shadow.innerHTML = """
<style>        
:host {
        display: inline-block;
        position: relative;
        font-family: Arial, sans-serif;
    }
    input {
        width: 90%;
        padding: 5px;
        background-color: #2a2a2a;
        color: #e0e0e0;
        border: 1px solid #444;
        border-radius: 4px;
        font-size: 14px;
    }
    input::placeholder {
        color: #888;
    }
    .dropdown {
        position: absolute;
        width: 100%;
        max-height: 200px;
        resize: both;
        overflow-y: auto;
        border: 1px solid #444;
        background-color: #333;
        color: #e0e0e0;
        border-radius: 0 0 4px 4px;
        box-shadow: 0 4px 6px rgba(0, 0, 0, 0.3);
    }
    .popup {
        position: absolute;
        z-index: 1000;
        display: none;
    }   
    .dropdown * {
        padding: 8px 12px;
        cursor: pointer;
        transition: background-color 0.2s ease;
    }
    .dropdown div:hover {
        background-color: #444;
    }
    input:focus {
        outline: none;
        border-color: #666;
        box-shadow: 0 0 0 2px rgba(100, 100, 100, 0.3);
    }      
</style>
        <input data-name="_input">
        <wwwpy-searchable-combobox2-option-popup 
            data-name="option_popup" class="popup" style="display: none">
        </wwwpy-searchable-combobox2-option-popup>
        """
        self.option_popup.parent = self
        self.option_popup.search_placeholder = 'search...'

    @property
    def text_value(self) -> str:
        return self._input.value

    @text_value.setter
    def text_value(self, value: str):
        self._input.value = value

    @property
    def placeholder(self) -> str:
        return self._input.placeholder

    @placeholder.setter
    def placeholder(self, value: str):
        self._input.placeholder = value

    def _input_element(self) -> js.HTMLElement:
        return self._input

    def _input__click(self, event):
        if len(self.option_popup.options) == 0:
            self._input.focus()
            return
        vis = self.option_popup.visible
        if vis:
            self.option_popup.hide()
        else:
            self.option_popup.show()
            if self.focus_search_on_popup:
                self.option_popup.focus_search()
            else:
                self._input.focus()

    def _input__change(self, event):
        self.element.dispatchEvent(js.CustomEvent.new('wp-change'))

    def _option_selected(self, option: Option):
        self.option_popup.hide()
        if option.actions.set_input_value:
            print(f'option selected: {option.text}')
            self._input.value = option.text
            self.element.dispatchEvent(js.CustomEvent.new('wp-change', dict_to_js({'detail': {'option': option}})))



import base64
import builtins
import enum
import importlib
import json
import re
import typing
from dataclasses import is_dataclass
from datetime import datetime
from typing import Any, Type, get_origin, get_args, TypeVar

T = TypeVar('T')

# todo this should also receive the expected cls Type.
# this way we can check that the instance is of the expected type
def serialize(obj: T, cls: Type[T]) -> Any:
    optional_type = _get_optional_type(cls)
    if optional_type:
        if obj is None:
            return None
        return serialize(obj, optional_type)
    origin = typing.get_origin(cls)
    if origin is typing.Union:
        args = set(get_args(cls))
        obj_type = type(obj)
        if obj_type not in args:
            raise ValueError(f"Expected object of type {args}, got {obj_type}")
        obj_ser = serialize(obj, obj_type)
        return [str(obj_type), obj_ser]

    if origin is not None:
        if not isinstance(obj, origin):
            raise ValueError(f"Expected object of type {origin}, got {type(obj)}")
    elif not isinstance(obj, cls):
        raise ValueError(f"Expected object of type {cls}, got {type(obj)}")
    if is_dataclass(obj):
        field_types = typing.get_type_hints(cls)
        return {
            field_name: serialize(getattr(obj, field_name), field_type)
            for field_name, field_type in field_types.items()
        }
    elif isinstance(obj, list):
        item_type = typing.get_args(cls)[0]
        return [serialize(item, item_type) for item in obj]
    elif isinstance(obj, tuple):
        return [serialize(item, get_args(cls)[i]) for i, item in enumerate(obj)]
    elif isinstance(obj, dict):
        key_type, value_type = typing.get_args(cls)
        return {
            serialize(key, key_type): serialize(value, value_type)
            for key, value in obj.items()
        }
    elif isinstance(obj, datetime):
        return obj.isoformat()
    elif isinstance(obj, bytes):
        return base64.b64encode(obj).decode('utf-8')
    elif isinstance(obj, (int, float, str, bool)):
        return obj
    elif isinstance(obj, enum.Enum):
        return serialize(obj.value, type(obj.value))
    elif obj is None:
        return None
    else:
        raise ValueError(f"Unsupported type: {type(obj)}")


def _get_optional_type(cls):
    origin = typing.get_origin(cls)
    args = typing.get_args(cls)

    if origin is typing.Union:
        if type(None) in args and len(args) == 2:
            return args[0]
    # if origin is types.UnionType:
    #     return type(None) in args
    return None


def deserialize(data: Any, cls: Type[T]) -> T:
    optional_type = _get_optional_type(cls)
    if optional_type:
        if data is None:
            return None
        return deserialize(data, optional_type)
    origin = get_origin(cls)
    if origin is typing.Union:
        args = set(get_args(cls))
        obj_type = _get_type_from_string(data[0])
        if obj_type not in args:
            raise ValueError(f"Expected object of type {args}, got {obj_type}")
        return deserialize(data[1], obj_type)
    if is_dataclass(cls):
        args = {}
        field_types = typing.get_type_hints(cls)
        for name, value in data.items():
            args[name] = deserialize(value, field_types[name])
        instance = cls(**args)
        return instance
    elif origin == list or cls == list:
        item_type = get_args(cls)[0]
        return [deserialize(item, item_type) for item in data]
    elif origin == tuple or cls == tuple:
        item_types = get_args(cls)
        return tuple(deserialize(data[i], item_types[i]) for i in range(len(data)))
    elif origin == dict or cls == dict:
        key_type, value_type = get_args(cls)
        return {
            deserialize(key, key_type): deserialize(value, value_type)
            for key, value in data.items()
        }
    elif issubclass(cls, list):  # for subclasses of list
        item_type = get_args(cls)[0]
        return cls([deserialize(item, item_type) for item in data])
    elif cls == datetime:
        return datetime.fromisoformat(data)
    elif cls == bytes:
        return base64.b64decode(data.encode('utf-8'))
    elif cls in (int, float, str, bool):
        return cls(data)
    elif issubclass(cls, enum.Enum):
        first_member = next(iter(cls))
        return cls(deserialize(data, type(first_member.value)))
    elif cls is type(None):
        return None
    else:
        raise ValueError(f"Unsupported type: {cls}")


def to_json(obj: Any, cls: Type[T]) -> str:
    return json.dumps(serialize(obj, cls))


def from_json(json_str: str, cls: Type[T]) -> T:
    data = json.loads(json_str)
    return deserialize(data, cls)


def _get_type_from_string(type_str):
    # Use regex to extract the full type path
    match = re.search(r"<class '(.+)'>|(.+)", type_str)
    if not match:
        raise ValueError(f"Invalid type string format: {type_str}")

    full_path = match.group(1) or match.group(2)

    if full_path == 'NoneType':
        return type(None)

    # Check if it's a builtin type
    if hasattr(builtins, full_path):
        return getattr(builtins, full_path)

    # Split the path into parts
    parts = full_path.split('.')

    # The class name is the last part
    class_name = parts.pop()

    # The module name is everything else
    module_name = '.'.join(parts)

    # Import the module dynamically
    module = importlib.import_module(module_name)

    # Get the class from the module
    return getattr(module, class_name)



from typing import List

from wwwpy.common.designer.element_library import ElementDef, Help, EventDef, AttributeDef


def _standard_elements_def() -> List[ElementDef]:
    res = [
        ElementDef(
            'button', 'js.HTMLButtonElement',
            help=Help('A clickable button.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/button'),
            attributes=[
                AttributeDef('type', Help('The type of the button.', ''),
                             values=['submit', 'reset', 'button'], default_value='button'),
                AttributeDef('value', Help('The value of the button.', '')),
            ],
            events=[EventDef('click', Help('The button was clicked.',
                                           'https://developer.mozilla.org/en-US/docs/Web/API/Element/click_event')),
                    EventDef('dblclick', Help('The button was double-clicked.',
                                              'https://developer.mozilla.org/en-US/docs/Web/API/Element/dblclick_event'))],
        ),
        ElementDef(
            'input', 'js.HTMLInputElement',
            help=Help('A field for entering text.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input'),
            attributes=[
                AttributeDef('type', Help('The type of control to display. The default type is text.', ''),
                             values=['text', 'password', 'checkbox', 'radio', 'button', 'submit', 'reset', 'file',
                                     'hidden', 'image', 'date', 'datetime-local', 'month', 'time', 'week',
                                     'number', 'range', 'email', 'url', 'search', 'tel', 'color'],
                             default_value='text'),
                AttributeDef('value', Help('The value of the input field.', '')),
                AttributeDef('placeholder', Help('A hint to the user of what can be entered in the field.', '')),
                AttributeDef('disabled', Help('Whether the control is disabled.', ''), boolean=True),
                AttributeDef('readonly', Help('Whether the control is read-only.', ''), boolean=True),
                AttributeDef('required', Help('Whether the control is required for form submission.', ''),
                             boolean=True),
                AttributeDef('min', Help('The minimum value of the input field.', '')),
                AttributeDef('max', Help('The maximum value of the input field.', '')),
                AttributeDef('step', Help('The legal number intervals for the input field.', '')),
                AttributeDef('pattern', Help('A regular expression that the input\'s value is checked against.', '')),
                AttributeDef('autocomplete', Help('Whether the control is required for form submission.', ''),
                             values=['on', 'off']),
                AttributeDef('autofocus', Help('Whether the control should have input focus when the page loads.', ''),
                             boolean=True),
                AttributeDef('checked', Help('Whether the control is checked.', ''), boolean=True),
                AttributeDef('multiple', Help('Whether the user is allowed to enter more than one value.', ''),
                             boolean=True),

            ],
            events=[
                EventDef('input', Help('The input event fires when the value of the element has been changed '
                                       'as a direct result of a user action',
                                       'https://developer.mozilla.org/en-US/docs/Web/API/Element/input_event')),
                EventDef('change')
            ]
        ),
        ElementDef(
            'div', 'js.HTMLDivElement',
            help=Help('A generic container element.',
                      'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/div')
        ),
        ElementDef(
            'progress', 'js.HTMLProgressElement',
            help=Help('A progress bar.',
                      'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/progress'),
            attributes=[
                AttributeDef('value', Help('The current value of the progress bar.', '')),
                AttributeDef('max', Help('The maximum value of the progress bar.', '')),
            ],
            events=[
                EventDef('click', Help('The progress bar was clicked.', '')),
                ])


    ]


    for r in res:
        r.gen_html = _generateHtml
    return res


def _generateHtml(element_def: ElementDef, name: str) -> str:
    tag_name = element_def.tag_name

    def _def(placeHolder=False, add=''):
        def inner():
            pl = '' if not placeHolder else f' placeholder="{name}"'
            add1 = '' if not add else f' {add}'
            return f'<{tag_name} data-name="{name}"{pl}{add1}>{name}</{tag_name}>'

        return inner

    func = {
        'button': _def(),
        'input': lambda: f"""<input data-name="{name}" placeholder="{name}">""",
        'progress': lambda: f"""<progress data-name="{name}" value="70" max="100">70 %</progress>""",
    }
    gen_html = func.get(tag_name, None)
    html = '\n' + gen_html() if gen_html else '' + ElementDef.default_gen_html(element_def, name)
    return html

# [
#     _comp('input', 'js.HTMLInputElement', '<input type="text" data-name="#name#" value="#name#">'),
#     _comp('button', 'js.HTMLButtonElement', '<button data-name="#name#">#name#</button>'),
#     _comp('textarea', 'js.HTMLTextAreaElement', '<textarea data-name="#name#">#name#</textarea>'),
#     _comp('select', 'js.HTMLSelectElement',
#           '<select data-name="#name#"><option>#name#</option></select>'),
#     _comp('div', 'js.HTMLDivElement', '<div data-name="#name#">#name#</div>'),
#     _comp('p', 'js.HTMLParagraphElement', '<p data-name="#name#">#name#</p>'),
#     _comp('h1', 'js.HTMLHeadingElement', '<h1 data-name="#name#">#name#</h1>'),
#     _comp('h2', 'js.HTMLHeadingElement', '<h2 data-name="#name#">#name#</h2>'),
#     _comp('h3', 'js.HTMLHeadingElement', '<h3 data-name="#name#">#name#</h3>'),
#     _comp('a', 'js.HTMLAnchorElement', '<a href="#" data-name="#name#">#name#</a>'),
#     _comp('img', 'js.HTMLImageElement', '<img src="#" alt="#name#" data-name="#name#">'),
#     _comp('ul', 'js.HTMLUListElement', '<ul data-name="#name#"><li>#name#</li></ul>'),
#     _comp('ol', 'js.HTMLOListElement', '<ol data-name="#name#"><li>#name#</li></ol>'),
#     _comp('li', 'js.HTMLLIElement', '<li data-name="#name#">#name#</li>'),
#     _comp('table', 'js.HTMLTableElement', '<table data-name="#name#"><tr><td>#name#</td></tr></table>'),
#     _comp('form', 'js.HTMLFormElement', '<form data-name="#name#">#name#</form>'),
#     _comp('label', 'js.HTMLLabelElement', '<label data-name="#name#">#name#</label>'),
#
# ]



from typing import Iterable, TypeVar, Generic, Callable, Iterator

_T = TypeVar('_T')


class repeatable_chain(Generic[_T]):
    def __init__(self, *iterables: Iterable[_T]):
        self._iterables = iterables

    def __iter__(self) -> Iterator[_T]:
        def it() -> Iterator[_T]:
            for iterable in self._iterables:
                yield from iterable

        return it()


class CallableToIterable(Generic[_T]):
    def __init__(self, iterator_factory: Callable[[], Iterator[_T]]):
        self.iterator_factory = iterator_factory

    def __iter__(self) -> Iterator[_T]:
        try:
            return iter_catching(self.iterator_factory())
        except:
            import traceback
            traceback.print_exc()
            return iter([])


def iter_catching(gen):
    while True:
        try:
            yield next(gen)
        except StopIteration:
            break
        except Exception as e:
            import traceback
            traceback.print_exc()



from js import window
from pyodide.ffi import create_once_callable
import asyncio


def pytest_sessionstart(session):
    print(f'invocation_dir={session.config.invocation_dir}')
    print(f'rootpath={session.config.rootpath}')


def pytest_xvirt_send_event(event_json):
    async def callback():
        path = '#xvirt_notify_path_marker#'
        await async_fetch_str(path, method='POST', data=event_json)

    asyncio.create_task(callback())


async def async_fetch_str(url: str, method: str = 'GET', data: str = '') -> str:
    print(f'url={url}')
    print(f'method={method}')
    print(f'data=`{data}`')
    response = await window.fetch(url, method=method, body=data)
    text = await response.text()
    return text




from dataclasses import dataclass
from typing import Callable, List
from contextlib import contextmanager


@dataclass
class PropertyChanged:
    instance: any
    name: str
    old_value: object
    new_value: object


def monitor_changes(instance, on_changed: Callable[[List[PropertyChanged]], None]):
    """Monitor the changes of the properties of an instance of a class."""
    if hasattr(instance, "__attr_change_monitor_on_changed"):
        raise Exception("The instance is already being monitored for property changes")

    clazz = instance.__class__
    if not hasattr(clazz, "__attr_change_monitor"):
        setattr(clazz, "__attr_change_monitor", True)

        original_setattr = clazz.__setattr__  # Keep a reference to the original method

        def new_setattr(self, name, value):
            doit = (hasattr(self, name) and name != "__attr_change_monitor_on_changed" and
                    hasattr(self, "__attr_change_monitor_on_changed"))
            if doit:
                change = PropertyChanged(self, name, getattr(self, name), value)
            original_setattr(self, name, value)
            if doit:
                self.__attr_change_monitor_on_changed([change])

        clazz.__setattr__ = new_setattr

    instance.__attr_change_monitor_on_changed = on_changed


@contextmanager
def group_changes(instance):
    old = instance.__attr_change_monitor_on_changed
    buffer = []

    def _in_buffer(changes):
        buffer.extend(changes)

    instance.__attr_change_monitor_on_changed = _in_buffer
    try:
        yield instance
    finally:
        instance.__attr_change_monitor_on_changed = old
        if buffer:
            old(buffer)



import logging
import warnings
from typing import Callable


class _CustomHandler(logging.Handler):
    def __init__(self, callback: Callable[[str], None]):
        super().__init__()
        self._emit = callback

    def emit(self, record: logging.LogRecord):
        log_entry = self.format(record)
        self._emit(log_entry)


def add_once(emit: Callable[[str], None]):
    for log_name in ['common', 'remote', 'server']:
        logging.getLogger(log_name).setLevel(logging.DEBUG)

    for log_name in ['wwwpy']:
        logging.getLogger(log_name).setLevel(logging.INFO)

    root = logging.getLogger()

    for handler in root.handlers:
        if isinstance(handler, _CustomHandler):
            return

    formatter = logging.Formatter('%(asctime)s %(levelname).1s %(name)s:%(lineno)d - %(message)s')
    custom_handler = _CustomHandler(emit)
    custom_handler.setFormatter(formatter)
    root.addHandler(custom_handler)


_original_showwarning = None


def _custom_showwarning(message, category, filename, lineno, file=None, line=None):
    if _original_showwarning:
        _original_showwarning(message, category, filename, lineno, file, line)
    log_entry = f'{category.__name__}: {message} ({filename}:{lineno})'
    logging.getLogger('common').warning(log_entry)


def warning_to_log():
    global _original_showwarning
    if _original_showwarning:
        return
    _original_showwarning = warnings.showwarning
    warnings.showwarning = _custom_showwarning



from pathlib import Path
from typing import Dict, Any
import shutil

from wwwpy.common.filesystem.sync import Event


def event_revert(path: Path, event: Event, file_contents: Dict[str, str]) -> Path:
    src_full_path = path / event.src_path
    dest_full_path = path / event.dest_path if event.dest_path else None

    if event.event_type == 'created':
        if event.is_directory:
            shutil.rmtree(src_full_path)
        else:
            src_full_path.unlink()

    elif event.event_type == 'deleted':
        if event.is_directory:
            src_full_path.mkdir(parents=True, exist_ok=True)
        else:
            src_full_path.touch()
            if event.src_path in file_contents:
                src_full_path.write_text(file_contents[event.src_path])

    elif event.event_type == 'modified':
        if not event.is_directory:
            if event.src_path in file_contents:
                src_full_path.write_text(file_contents[event.src_path])
            else:
                # If we don't have the previous content, we can't revert properly
                # For now, we'll just empty the file
                src_full_path.write_text('')

    elif event.event_type == 'moved':
        if dest_full_path.exists():
            if event.is_directory:
                shutil.move(str(dest_full_path), str(src_full_path))
            else:
                dest_full_path.rename(src_full_path)

    # For 'closed' event, we don't need to do anything

    return path


import random

adjectives = [
    'adaptive', 'agile', 'ambitious', 'analytical', 'bold', 'brilliant', 'clever', 'crazy', 'creative', 'curious',
    'daring', 'determined', 'dynamic', 'efficient', 'energetic', 'fearless', 'focused', 'innovative', 'insightful',
    'inspired', 'intuitive', 'masterful', 'mighty', 'nimble', 'passionate', 'proactive', 'quick', 'reliable',
    'resourceful', 'savvy', 'sharp', 'skillful', 'sleek', 'smart', 'strategic', 'talented', 'tenacious',
    'versatile', 'visionary', 'witty'
]

roles = [
    'ace', 'analyst', 'architect', 'artisan', 'builder', 'champion', 'coder', 'connoisseur', 'creator', 'debugger',
    'designer', 'developer', 'dynamo', 'engineer', 'expert', 'genius', 'guru', 'innovator', 'luminary', 'maestro',
    'maintainer', 'manager', 'master', 'maverick', 'ninja', 'operator', 'pioneer', 'prodigy', 'programmer', 'sage',
    'specialist', 'strategist', 'technician', 'trailblazer', 'tuner', 'virtuoso', 'visionary', 'whiz', 'wizard',
    'wrangler'
]


def generate_name():
    return f"{random.choice(adjectives)}-{random.choice(roles)}"



import traceback
from inspect import getmembers, isfunction, signature, iscoroutinefunction, Signature
from types import ModuleType, FunctionType
from typing import NamedTuple, List, Tuple, Any, Optional, Dict, Callable, Awaitable, Protocol, Iterator

from wwwpy.common import modlib
from wwwpy.common.iterlib import CallableToIterable
from wwwpy.common.rpc.serializer import RpcRequest, RpcResponse
from wwwpy.exceptions import RemoteException
from wwwpy.http import HttpRoute, HttpResponse, HttpRequest
from wwwpy.resources import Resource, StringResource, ResourceIterable
from wwwpy.unasync import unasync


class Function(NamedTuple):
    name: str
    func: FunctionType
    signature: str
    is_coroutine_function: bool
    sign: Signature
    blocking: FunctionType


def _std_function_to_function(fun_tuple: Tuple[str, FunctionType]) -> Function:
    name = fun_tuple[0]
    func = fun_tuple[1]
    sign = signature(func)
    is_coroutine_function = iscoroutinefunction(func)
    blocking = unasync(func)
    return Function(name, func, str(sign), is_coroutine_function, signature(func), blocking)


class Module:
    def __init__(self, module: ModuleType):
        self.module = module
        self.name = module.__name__

        self.functions: List[Function] = function_list(self.module)
        self._funcs = {f.name: f for f in self.functions}

    def __getitem__(self, name) -> Function:
        return self._funcs.get(name, None)


def function_list(module: ModuleType) -> List[Function]:
    return list(map(_std_function_to_function, getmembers(module, isfunction)))


Fetch = Callable[[str, str, str], Awaitable[str]]


class Fetch(Protocol):
    def __call__(self, url: str, method: str = '', data: str = '') -> Awaitable[str]: ...


class Proxy:
    def __init__(self, module_name: str, rpc_url: str, fetch: Fetch):
        self.rpc_url = rpc_url
        self.fetch = fetch
        self.module_name = module_name

    async def dispatch_async(self, func_name: str, *args) -> Any:
        rpc_request = RpcRequest.build_request(self.module_name, func_name, *args)
        json_response = await self.fetch(self.rpc_url, method='POST', data=rpc_request.json())
        response = RpcResponse.from_json(json_response)
        ex = response.exception
        if ex is not None and ex != '':
            raise RemoteException(ex)
        return response.result

    def dispatch_sync(self, func_name: str, *args) -> Any:
        rpc_request = RpcRequest.build_request(self.module_name, func_name, *args)
        import js
        xhr = js.XMLHttpRequest.new()
        xhr.open('POST', self.rpc_url, False)
        xhr.setRequestHeader('Content-Type', 'application/json')
        xhr.send(rpc_request.json())
        json_response = xhr.responseText
        response = RpcResponse.from_json(json_response)
        ex = response.exception
        if ex is not None and ex != '':
            raise RemoteException(ex)
        return response.result


class RpcRoute:
    def __init__(self, route_path: str):
        self._allowed_modules: set[str] = set()
        self.route = HttpRoute(route_path, self._route_callback)

    def _route_callback(self, request: HttpRequest) -> HttpResponse:
        resp = self.dispatch(request.content)
        response = HttpResponse(resp, 'application/json')
        return response

    def add_module(self, module_name: str):
        # todo rename to `allow`
        if not isinstance(module_name, str):
            raise TypeError('module_name must be a string')
        self._allowed_modules.add(module_name)

    def find_module(self, module_name: str) -> Optional[Module]:
        if module_name not in self._allowed_modules:
            return None
        if modlib._find_module_path(module_name) is None:
            return None
        import importlib
        module = importlib.import_module(module_name)
        # todo, cache? beware of hot reload
        return Module(module)

    def dispatch(self, request: str) -> str:
        rpc_request = RpcRequest.from_json(request)
        # print(f'dispatch req={request}')
        module = self.find_module(rpc_request.module)
        function = module[rpc_request.func]
        exception = ''
        result = None
        try:
            result = function.blocking(*rpc_request.args)
        except Exception:
            exception = traceback.format_exc()

        response = RpcResponse(result, exception)
        return response.to_json()

    def remote_stub_resources(self) -> ResourceIterable:

        def bundle() -> Iterator[Resource]:
            for module_name in self._allowed_modules:
                module = self.find_module(module_name)
                if module is None:
                    continue
                imports = 'from wwwpy.remote.fetch import async_fetch_str'
                stub_source = generate_stub_source(module, self.route.path, imports)
                yield StringResource(module_name.replace('.', '/') + '.py', stub_source)

        return CallableToIterable(bundle)


def generate_stub_source(module: Module, rpc_url: str, imports: str):
    module_name = module.name
    # language=python
    stub_header = f"""
from __future__ import annotations
from wwwpy.rpc import Proxy
{imports}

rpc_url = '{rpc_url}'
module_name = '{module_name}'
proxy = Proxy(module_name, rpc_url, async_fetch_str)
    """

    stub_functions = ''
    for f in module.functions:
        if f.name.startswith('_'):
            continue
        parameters = f.sign.parameters.values()
        params_list = ', '.join(p.name for p in parameters)
        args_list = '' if params_list == '' else ', ' + params_list
        fun_stub_as = f'\nasync def {f.name}{f.signature}:\n' + \
                      f'    return await proxy.dispatch_async("{f.name}"{args_list})\n'
        fun_stub_sy = f'\ndef {f.name}{f.signature}:\n' + \
                      f'    return proxy.dispatch_sync("{f.name}"{args_list})\n'
        stub_functions += fun_stub_as if f.is_coroutine_function else fun_stub_sy

    return stub_header + stub_functions



from __future__ import annotations

import ast
import importlib.util
from ast import Module, FunctionDef, ClassDef
from typing import NamedTuple, List


class Function(NamedTuple):
    name: str
    signature: str
    is_coroutine_function: bool
    parameters: List[Parameters]


class Parameters(NamedTuple):
    name: str
    annotation: str


class FuncMeta:
    """It contains only the information about the functions of a module.
    It does not have the actual functions.
    """

    def __init__(self, name: str, functions: List[Function]):
        self.name = name

        self.functions: List[Function] = functions
        self._funcs = {f.name: f for f in self.functions}

    def __getitem__(self, name) -> Function:
        return self._funcs.get(name, None)


def from_package_name(package_name: str) -> FuncMeta | None:
    """it returns a FuncRegistry object from the given package name. If the package is not found, it returns None.
    The package will not be loaded it will be parsed with ast module.
    """
    spec = importlib.util.find_spec(package_name)
    if spec is None:
        return None

    source_code = spec.loader.get_source(package_name)
    functions = function_definitions(source_code)
    return FuncMeta(package_name, functions)


def function_definitions(source_code) -> List[Function]:
    tree = ast.parse(source_code)
    functions = []

    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            function_name = node.name
            args = []
            parameters = []
            for arg in node.args.args:
                arg_name = arg.arg
                arg_annotation = ast.get_source_segment(source_code, arg.annotation) if arg.annotation else None
                args.append(f"{arg_name}: {arg_annotation}" if arg_annotation else arg_name)
                parameters.append(Parameters(arg_name, arg_annotation))

            returns = ast.get_source_segment(source_code, node.returns) if node.returns else None
            signature = '(' + ', '.join(args) + ')' + (f' -> {returns}' if returns else '')
            f = Function(function_name, signature, isinstance(node, ast.AsyncFunctionDef), parameters)
            functions.append(f)

    return functions


def source_to_proxy(module_name: str, source: str):
    tree: Module = ast.parse(source)
    content = ''
    # todo when detecting unsupported structure (e.g, nested class, AsyncFunctionDef, functions with results), log a warning message

    for b in tree.body:
        if isinstance(b, ClassDef):
            content += f'class {b.name}:\n' \
                       '    def __init__(self, send_endpoint):\n' \
                       '        self.send_endpoint = send_endpoint\n'

            for f in b.body:
                fqn = b.name + '.' + f.name
                if isinstance(f, FunctionDef):
                    if len(f.args.args) > 1:  # beyond self
                        args = ', ' + ', '.join([a.arg for a in f.args.args[1:]])
                    else:
                        args = ''
                    content += f'    def {f.name}(self{args}):\n' \
                               f'       self.send_endpoint.dispatch("{module_name}", "{fqn}"{args})\n'

    return content



from __future__ import annotations

import js

import wwwpy.remote.component as wpc
from wwwpy.remote import dict_to_js


class HelpIcon(wpc.Component, tag_name='wwwpy-help-icon'):
    _link: js.HTMLElement = wpc.element()
    href: str = wpc.attribute()

    def root_element(self):
        return self.shadow

    def init_component(self):
        self.shadow = self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.shadow.innerHTML = """
<style>        
 .help-icon {
            width: 18px;
            height: 18px;
            vertical-align: middle;
            margin-left: 5px;
        }
</style>        
<svg xmlns="http://www.w3.org/2000/svg" style="display: none;">
    <symbol id="help-icon" viewBox="0 0 18 18">
        <circle cx="9" cy="9" r="8" fill="#007bff"/>
        <text x="9" y="9" fill="white" text-anchor="middle" dominant-baseline="central" font-weight="bold" font-size="12">?</text>
    </symbol>
</svg>   
<a data-name="_link" style="text-decoration: none" target="_blank">
<svg class="help-icon"><use href="#help-icon"/></svg></a>
"""
        self._update_href()

    def _link__click(self, event: js.MouseEvent):
        event.stopPropagation()

    def attributeChangedCallback(self, name: str, oldValue: str, newValue: str):
        if name == 'href':
            self._update_href()

    def _update_href(self):
        self._link.href = self.href




from threading import Thread
from typing import Optional

import uvicorn
from fastapi import FastAPI
from fastapi.responses import Response

from wwwpy.http import HttpRoute, HttpResponse, HttpRequest
from ..webserver import Webserver, wait_forever


class WsFastapi(Webserver):
    def __init__(self):
        super().__init__()
        self.app = FastAPI()
        self.thread: Optional[Thread] = None

    def _setup_route(self, route: HttpRoute):
        async def func(req) -> Response:
            if req.method == 'POST':
                body = await req.body()
                wwwpy_req = HttpRequest(req.method, body, req.headers)
            else:
                wwwpy_req = HttpRequest(req.method, req.body, req.headers)
            return self.to_native_response(route.callback(wwwpy_req))

        self.app.add_route(route.path, route=func, methods=['GET','POST'])

    def to_native_response(self, pn_response: HttpResponse):
        return Response(content=pn_response.content, media_type=pn_response.content_type)

    def _start_listen(self):
        def run():
            uvicorn.run(self.app, host=self.host, port=self.port)

        self.thread = Thread(target=run, daemon=True)
        self.thread.start()


if __name__ == '__main__':
    s = WsFastapi()
    s.set_http_route(HttpRoute('/', lambda req: HttpResponse('ciao', 'text/html')))
    s.start_listen()
    wait_forever()



class CustomStr(str):
    pass


def get_root_folder_or_fail() -> CustomStr:
    import sys
    only_custom = [p for p in sys.path if isinstance(p, CustomStr)]

    if len(only_custom) != 1:
        raise ValueError('Cannot find root folder')

    return only_custom[0]



from __future__ import annotations

import sys

_wwwpy_dev_mode = 'wwwpy_dev_mode'


async def activate():
    setattr(sys, _wwwpy_dev_mode, True)
    from wwwpy.remote import micropip_install
    from wwwpy.common import designer

    for package in designer.pypi_packages:
        await micropip_install(package)

def is_active() -> bool:
    return hasattr(sys, _wwwpy_dev_mode)



from js import document
import js

def setup_shoelace():
    document.documentElement.className = 'sl-theme-dark'
    document.head.append(document.createRange().createContextualFragment(_head_style))



# language=html
_head_style = """
<meta name="viewport" content="width=device-width, initial-scale=1.0, maximum-scale=1.0, user-scalable=0">
<style>@import 'https://cdn.jsdelivr.net/npm/@shoelace-style/shoelace@2.15.1/cdn/themes/dark.css';</style>
<style>
    body {
        font: 16px sans-serif;
        background-color: var(--sl-color-neutral-0);
        color: var(--sl-color-neutral-900);
        padding: 1rem;
    }
</style>
<script type="module" src="https://cdn.jsdelivr.net/npm/@shoelace-style/shoelace@2.15.1/cdn/shoelace.js"></script>
"""



from tornado.web import RequestHandler

GLOBAL_CHARSET = "utf-8"


class AsgiHandler(RequestHandler):

    def initialize(self, asgi_app) -> None:
        super().initialize()
        self._asgi_app = asgi_app

    async def handle_request(self):
        headers = []
        for k in self.request.headers:
            for v in self.request.headers.get_list(k):
                headers.append(
                    (k.encode(GLOBAL_CHARSET).lower(), v.encode(GLOBAL_CHARSET))
                )

        scope = {
            "type": self.request.protocol,
            "http_version": self.request.version,
            "path": self.request.path,
            "method": self.request.method,
            "query_string": self.request.query.encode(GLOBAL_CHARSET),
            "headers": headers,
            "client": (self.request.remote_ip, 0)
        }

        async def receive():
            return {'body': self.request.body, "type": "http.request", "more_body": False}

        async def send(data):
            if data['type'] == 'http.response.start':
                self.set_status(data['status'])
                self.clear_header("content-type")
                self.clear_header("server")
                self.clear_header("date")
                for h in data['headers']:
                    if len(h) == 2:
                        self.add_header(
                            h[0].decode(GLOBAL_CHARSET),
                            h[1].decode(GLOBAL_CHARSET)
                        )
            elif data['type'] == 'http.response.body':
                self.write(data['body'])
            else:
                raise RuntimeError(
                    f"Unsupported response type \"{data['type']}\" for asgi app")

        await self._asgi_app(scope, receive, send)

    async def get(self):
        await self.handle_request()

    async def post(self):
        await self.handle_request()